import fs from 'fs'
import path from 'path'
import ora from 'ora'
import kebabCase from 'lodash.kebabcase'
import url from 'url'
import { print } from '@toptal/davinci-cli-shared'
import { createRequire } from 'module'

import workspace from '../utils/workspace.js'
// eslint-disable-next-line no-restricted-syntax
import skeleton from '../utils/skeleton.js'
import { APPS_PATH, NAMESPACES_PATH } from '../constants.js'
const localRequire = createRequire(import.meta.url)
const thisScriptDirname = path.dirname(url.fileURLToPath(import.meta.url))
const getAppPath = (libName, namespace) =>
  namespace
    ? path.join(NAMESPACES_PATH, namespace, APPS_PATH, libName)
    : path.join(APPS_PATH, libName)
const checkPrerequisites = async (monorepoWorkspaceDir, appName, namespace) => {
  const spinner = ora('Checking prerequisites...').start()

  // check if run in monorepo
  const isWorkspace = Boolean(monorepoWorkspaceDir)

  if (!isWorkspace) {
    spinner.fail()
    throw new Error(
      "It looks like you run 'new-app' Davinci command not in a monorepo."
    )
  }

  // check if app already exists
  const appPath = path.join(
    monorepoWorkspaceDir,
    getAppPath(appName, namespace)
  )
  const doesAppExist = fs.existsSync(appPath)

  if (doesAppExist) {
    spinner.fail()
    throw new Error(`Application already exists in the folder ${appPath}!`)
  }
  spinner.succeed('Prerequisites checked!')
}
const copyApplicationSkeleton = async (destDir, currentDavinciRunningDir) => {
  const skeletonSubAppPath = localRequire.resolve(
    '@toptal/davinci-skeleton-sub-app',
    {
      paths: [currentDavinciRunningDir],
    }
  )
  const appDir = path.join(skeletonSubAppPath, '../..')
  const spinner = ora('Copying skeleton...').start()

  try {
    await skeleton.copyAllFiles(appDir, destDir)
    await skeleton.copyTemplatePackageJsonFile(destDir)
    spinner.succeed('Skeleton copied!')
  } catch (err) {
    spinner.fail()
    throw err
  }
}
const modifyPackageJson = async (
  destDir,
  appName,
  { namespace, workspacePackageName }
) => {
  const spinner = ora('Modifying package.json...').start()

  try {
    const config = skeleton.readPackageJson(destDir)

    config.name = namespace
      ? `@toptal/${namespace}-${appName}`
      : `@toptal/${appName}`
    config.homepage = `https://github.com/toptal/${workspacePackageName}/tree/master/${getAppPath(
      appName,
      namespace
    )}`
    delete config.gitHead
    delete config.publishConfig
    config.license = 'SEE LICENSE IN LICENSE.MD'
    config.version = '0.0.1'
    skeleton.writePackageJson(destDir, config)
    spinner.succeed('package.json modified!')
  } catch (err) {
    spinner.fail()
    throw err
  }
}
const installDependencies = async monorepoWorkspaceDir => {
  const spinner = ora('Installing dependencies... ').start()

  try {
    await workspace.installDependencies(monorepoWorkspaceDir)
    spinner.succeed('Dependencies installed!')
  } catch (err) {
    spinner.fail()
    throw new Error(err)
  }
}
const createCommand = async (appName, namespace) => {
  print.header(
    'Creating',
    appName,
    namespace ? `in namespace ${namespace}` : '',
    '...'
  )
  const monorepoWorkspaceDir = workspace.getRootDir()
  const workspacePackageJson = workspace.readPackageJson(monorepoWorkspaceDir)
  const currentDavinciRunningDir = thisScriptDirname

  try {
    // check if all apps are installed, directory doesn't exists, etc.
    await checkPrerequisites(monorepoWorkspaceDir, appName, namespace)
    const destDir = path.join(
      monorepoWorkspaceDir,
      getAppPath(appName, namespace)
    )
    const workspacePackageName = workspace.getPackageName(workspacePackageJson)

    // copy skeleton to destination directory
    await copyApplicationSkeleton(destDir, currentDavinciRunningDir)

    // change package.json
    await modifyPackageJson(destDir, appName, {
      namespace,
      workspacePackageName,
    })

    // run yarn
    await installDependencies(monorepoWorkspaceDir)
    print.success('Application has been created.')
  } catch (err) {
    print.red(err)
    process.exit(1)
  }
}

export const createNewAppCommand = program => {
  return program
    .createCommand('new-app')
    .description('Create new application in monorepo')
    .action(async (appName, { namespace }) => {
      const formattedAppName = kebabCase(appName)

      return createCommand(formattedAppName, namespace)
    })
    .argument(
      '<appName>',
      'Application Name: folder name for the new application'
    )
    .option(
      '--namespace <namespace>',
      "Namespace of the app. Creates the app in a monorepo namespace: './namespaces/{namespace}/apps/{appName}'"
    )
}
